/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2008
   derniere modification le 15/09/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECore.h>
#include <SCE/interface/SCEScene.h>

#include <SCE/utils/SCEVector.h>
#include <SCE/interface/SCEBoundingBox.h>


/**
 * \file SCEScene.c
 * \copydoc scene
 * 
 * \file SCEScene.h
 * \copydoc scene
 */

/**
 * \defgroup scene Scene managment
 * \ingroup interface
 * \brief Scene managment
 * \internal
 */

/** @{ */

static SCE_SCamera *default_camera = NULL;

/* revise le 16/09/2008 */
/** \internal */
int SCE_Init_Scene (void)
{
    if (!(default_camera = SCE_Camera_Create ()))
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}
/* revise le 16/09/2008 */
/** \internal */
void SCE_Quit_Scene (void)
{
    SCE_Camera_Delete (default_camera);
    default_camera = NULL;
}


/* ajoute le 10/07/2008 */
/* ajoute le 15/07/2008 */
static void SCE_Scene_InitRenderStates (SCE_SSceneRenderStates *states)
{
    /* parametres par defaut des scenes */
    states->frustum_culling = SCE_FALSE;
    states->use_skybox = SCE_FALSE;
    states->lighting = SCE_TRUE;
    states->lod = SCE_FALSE;
}

/* ajoute le 03/04/2008 */
/* revise le 10/07/2008 */
static void SCE_Scene_InitRenderInfos (SCE_SSceneRenderInfos *infos)
{
    infos->rendertarget = NULL;
    infos->cubeface = 0;
    infos->camera = NULL;
    infos->models = NULL;
    SCE_Scene_InitRenderStates (&infos->states);
}

/* ajoute le 23/02/2008 */
/* revise le 06/04/2008 */
static void SCE_Scene_Init (SCE_SScene *scene)
{
    int i;
    scene->rootnode = NULL;
    scene->models = NULL;
    scene->lights = NULL;
    for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
        SCE_Scene_InitRenderInfos (&scene->render[i]);
    scene->ri = 0;
}

static void SCE_Scene_DeleteModelRenderInfos (void *infos)
{
    SCE_free (infos);
}

/* ajoute le 23/02/2008 */
/* revise le 20/03/2008 */
/**
 * \brief Creates a new scene
 * \returns a newly allocated scene or NULL on error
 */
SCE_SScene* SCE_Scene_Create (void)
{
    int i;
    SCE_SScene *scene = NULL;
    SCE_btstart ();
    if (!(scene = SCE_malloc (sizeof *scene)))
        Logger_LogSrc ();
    else
    {
        SCE_Scene_Init (scene);
#define SCE_SCENE_ASSERT(c)\
        if (c)\
        {\
            SCE_Scene_Delete (scene);\
            Logger_LogSrc ();\
            SCE_btend ();\
            return NULL;\
        }

        SCE_SCENE_ASSERT (!(scene->models = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->lights = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->rootnode = SCE_Node_Create ()))
        for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
        {
            scene->render[i].models =
                SCE_List_Create (SCE_Scene_DeleteModelRenderInfos);
            SCE_SCENE_ASSERT (!(scene->render[i].models))
        }
#undef SCE_SCENE_ASSERT
    }
    SCE_btend ();
    return scene;
}
/* ajoute le 23/02/2008 */
/**
 * \brief Deletes a scene
 * \param scene the scene you want to delete
 */
void SCE_Scene_Delete (SCE_SScene *scene)
{
    if (scene)
    {
        unsigned int i;
        for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
            SCE_List_Delete (scene->render[i].models);
        SCE_Node_Delete (scene->rootnode);
        SCE_List_Delete (scene->lights);
        SCE_List_Delete (scene->models);
        SCE_free (scene);
    }
}


/* ajoute le 20/03/2008 */
/**
 * \brief Generates a bouding box for each scene model
 * \param scene a scene
 * 
 * This function generates the bouding box by calling
 * SCE_Model_GenerateBoundingBox() for each model in the given scene.
 * 
 * \see SCE_Model_GenerateBoundingBox()
 */
void SCE_Scene_GenerateBoundingsBoxes (SCE_SScene *scene)
{
    SCE_SListIterator *it = NULL;
    SCE_List_ForEach (it, scene->models)
        SCE_Model_GenerateBoundingBox (SCE_List_GetData (it));
}


/* ajoute le 26/02/2008 */
/**
 * \brief Gets the root node of a scene
 * \param scene a scene
 * \returns the scene's root node
 */
SCE_SNode* SCE_Scene_GetRootNode (SCE_SScene *scene)
{
    return scene->rootnode;
}

/* ajoute le 26/02/2008 */
/**
 * \brief Adds a node to a scene
 * \param scene a scene
 * \param node the nod you want to attach to \p scene
 * \returns SCE_OK on succes or SCE_ERROR on falure.
 * 
 * This function attach the given note to the root node of the given scene. It
 * is equivalent to SCE_Node_Attach(SCE_Scene_GetRootNode(\p scene), \p node)
 * 
 * \see SCE_Node_Attach()
 */
int SCE_Scene_AddNode (SCE_SScene *scene, SCE_SNode *node)
{
    if (SCE_Node_Attach (scene->rootnode, node) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 26/02/2008 */
/* revise le 08/07/2008 */
/**
 * \brief Adds a model to a scene
 * \param scene a scene
 * \param model a model to add to \p scene
 * \returns SCE_ERROR on error, SCE_OK otherwise
 */
int SCE_Scene_AddModel (SCE_SScene *scene, SCE_SModel *model)
{
    int i;

    /* ajout du modele a la liste des modeles */
    if (SCE_List_AddBack (scene->models, model) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }

    /* ajout des elements aux listes */
    for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
    {
        SCE_SSceneModelRenderInfos *infos = SCE_malloc (sizeof *infos);
        if (!infos)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        infos->model = model;
        infos->num_lights = 0;
        /* ajout de l'infos de rendu pour le modele */
        if (SCE_List_AddBack (scene->render[i].models, infos) < 0)
        {
            Logger_LogSrc ();
            SCE_free (infos); /* TODO: ne frees pas les precedents... osef */
            return SCE_ERROR;
        }
    }
    return SCE_OK;
}

/* ajoute le 06/04/2008 */
/* revise le 07/04/2008 */
/**
 * \brief Adds a skybox model to a scene
 * \param scene a scene
 * \param skybox a model that will be added as a skybox
 * \returns SCE_ERROR on error, SCE_OK otherwise
 * 
 * This function adds a skybox model to the given scene.
 */
int SCE_Scene_AddSkybox (SCE_SScene *scene, SCE_SModel *skybox)
{
    if (SCE_List_AddFront (scene->models, skybox) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    if (!SCE_Model_HaveRenderFunc (skybox))
        SCE_Model_SetRenderFunc (skybox, SCE_Scene_RenderSkybox, scene);
    return SCE_OK;
}

/* ajoute le 20/03/2008 */
/**
 * \brief Adds a light to a scene
 * \param scene a scene
 * \param light a light to add to the given scene
 * \returns SCE_ERROR on error, SCE_OK otherwise
 */
int SCE_Scene_AddLight (SCE_SScene *scene, SCE_SLight *light)
{
    if (SCE_List_AddBack (scene->lights, light) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}


/* ajoute le 21/07/2008 */
/**
 * \brief Calls a function for each scene's models
 * \param scene a scene
 * \param func the function to call for each model
 * \param param user data to pass as second argument of \p func
 * 
 * \see SCE_FSceneForEachModelFunc
 */
void SCE_Scene_ForEachModel (SCE_SScene *scene, SCE_FSceneForEachModelFunc func,
                             void *param)
{
    SCE_SListIterator *it;
    SCE_List_ForEach (it, scene->models)
    {
        if (!func (SCE_List_GetData (it), param))
            break;
    }
}
/* ajoute le 22/07/2008 */
/**
 * \brief Calls a function for each scene's model that have been selected for
 *        render.
 * \param scene a scene
 * \param func the function to call for each selected model
 * \param param user data to pass as second argument of \p func
 * 
 * This function calls the given function for each model of the given scene that
 * have been selected for the render.
 * 
 * \sa SCE_FSceneForEachModelFunc SCE_Scene_ForEachModel()
 */
void SCE_Scene_ForEachSelectedModel (SCE_SScene *scene,
                                     SCE_FSceneForEachModelFunc func,
                                     void *param)
{
    unsigned int ri = scene->ri;
    SCE_SSceneModelRenderInfos *infos = NULL;
    SCE_SListIterator *it;
    SCE_List_ForEach (it, scene->render[ri].models)
    {
        infos = SCE_List_GetData (it);
        if (infos->draw)
        {
            if (!func (infos->model, param))
                break;
        }
    }
}


/* ajoute le 03/04/2008 */
/* revise le 16/09/2008 */
/* TODO: cette func aussi elle pue */
/**
 * \internal
 * \deprecated
 */
void SCE_Scene_RenderSkybox (SCE_SModel *mdl, void *s)
{
    SCE_SScene *scene = s;
    SCE_SCamera *cam = scene->render[scene->ri].camera;
    SCE_TVector3 campos = {0., 0., 0.};
    SCE_TMatrix4 cmat;
    float *mat = SCE_Node_GetMatrix (SCE_Model_GetNode (mdl));

    SCE_Matrix4_Copy (cmat, SCE_Camera_GetFinalView (cam));
    SCE_Matrix4_Inverse (cmat);
    SCE_Matrix4_GetTranslation (cmat, campos);
    SCE_Matrix4_Copy (cmat, mat); /* sauvegarde de la matrice */
    SCE_Matrix4_MulTranslatev (mat, campos);
    /* glPushAttribs */
    SCE_CSetState (GL_DEPTH_TEST, SCE_FALSE);
    SCE_Model_Render (mdl);
    SCE_CSetState (GL_DEPTH_TEST, SCE_TRUE);
    SCE_Matrix4_Copy (mat, cmat); /* restauration de la matrice */
}


/* ajoute le 05/04/2008 */
/* indique si un modele est dans le frustum de vue */
static int SCE_Scene_ModelInFrustum (SCE_SScene *scene, SCE_SCamera *cam,
                                     SCE_SModel *mdl)
{
    int render;
    SCE_SBoundingBox *box = SCE_Model_GetBoundingBox (mdl);

    /* application de la matrice a la bouding box */
    SCE_BoundingBox_Push (box, SCE_Node_GetFinalMatrix(SCE_Model_GetNode(mdl)));
    /* on verifie si le modele est dans le frustum */
    render = SCE_Frustum_BoundingBoxIn (SCE_Camera_GetFrustum (cam),
                                        SCE_Model_GetBoundingBox (mdl));
    SCE_BoundingBox_Pop (box);
    return render;
}

/* ajoute le 07/07/2008 */
/* construit la liste des modeles qui seront dessines */
static void SCE_Scene_SelectInFrustumModels (SCE_SScene *scene,
                                             SCE_SCamera *cam)
{
    SCE_SListIterator *i = NULL;
    SCE_SSceneModelRenderInfos *info = NULL;
    int ri = scene->ri;

    /* TODO: risque de faire des calculs doublons */
    SCE_Camera_UpdateFrustum (cam);

    /* pour chaque modele... */
    SCE_List_ForEach (i, scene->render[ri].models)
    {
        info = SCE_List_GetData (i);
        /* s'il a ete valide dans les (eventuels) precedents tests,
           on effectue ce nouveau test */
        if (info->draw)
        {
            /* ... on verifie s'il est dans le frustum de vue */
            if (!SCE_Scene_ModelInFrustum (scene, cam, info->model))
                info->draw = SCE_FALSE;
        }
    }
}

/* ajoute le 11/07/2008 */
/* revise le 21/07/2008 */
static void SCE_Scene_SelectActivedModels (SCE_SScene *scene)
{
    SCE_SListIterator *i = NULL;
    SCE_SSceneModelRenderInfos *info = NULL;
    int ri = scene->ri;

    /* pour chaque modele */
    SCE_List_ForEach (i, scene->render[ri].models)
    {
        info = SCE_List_GetData (i);
        info->draw = SCE_Model_IsActived (info->model);
    }
}

/* ajoute le 07/07/2008 */
/* revise le 16/07/2008 */
/* toutes les lumieres dont infos->model est a portee sont enregistrees */
static void SCE_Scene_MakeLightsList (SCE_SScene *scene,
                                      SCE_SSceneModelRenderInfos *infos)
{
    SCE_SListIterator *i = NULL;
    unsigned int j = 0;
    SCE_SLight *light = NULL;
    SCE_SBoundingBox *box = NULL;
    SCE_TVector3 vec1, vec2;
    float length;
    SCE_TVector3 p;
    SCE_SModel *mdl = infos->model;

    box = SCE_Model_GetBoundingBox (mdl);
    if (!box) /* TODO: ahem... */
        return;

    /* obtention de donnees necessaire au calcul des lumieres */
    SCE_BoundingBox_Push (box, SCE_Node_GetFinalMatrix(SCE_Model_GetNode(mdl)));
    SCE_BoundingBox_GetCenterv (box, vec1);
    SCE_Vector3_Copy (vec2, SCE_BoundingBox_GetPoints (box));
    SCE_BoundingBox_Pop (box);
    length = SCE_Vector3_Distance (vec1, vec2);

    /* activation des lumieres */
    SCE_List_ForEach (i, scene->lights)
    {
        light = SCE_List_GetData (i);

        /* on verifie d'abord si la lumiere est activee */
        if (!SCE_Light_IsActived (light))
            continue;

        /* on n'active la lumiere que si le modele est a sa portee */
        SCE_Light_CopyPosition (light, p);
        if (SCE_Light_IsInfinite (light) ||
            SCE_Vector3_Distance (p, vec1) <= SCE_Light_GetRadius(light)+length)
        {
            /* TODO: gerer les spot lights */
            /* ajout du modele a la liste des modeles */
            infos->lights[j] = light;
            j++;
            /* TODO: je ne considere que les premieres lumieres de la liste
               et non les plus importantes (cad les plus intenses/proches)
               penser a trier scene->lights, et ca sera cool */
            if (j >= SCE_MAX_LIGHTS_ON_MODEL)
                break;
        }
    }
    infos->num_lights = j;
}

/* ajoute le 08/07/2008 */
static void SCE_Scene_MakeLightsLists (SCE_SScene *scene)
{
    SCE_SSceneModelRenderInfos *info = NULL;
    SCE_SListIterator *i = NULL;
    int ri = scene->ri;

    if (scene->render[ri].states.lighting)
    {
        SCE_List_ForEach (i, scene->render[ri].models)
        {
            info = SCE_List_GetData (i);
            /* uniquement si le modele a passe la validation pour etre rendu */
            if (info->draw)
                SCE_Scene_MakeLightsList (scene, SCE_List_GetData (i));
        }
    }
    else
    {
        SCE_List_ForEach (i, scene->render[ri].models)
        {
            info = SCE_List_GetData (i);
            info->num_lights = 0;
        }
    }
}

/* ajoute le 15/07/2008 */
/* ajoute le 21/07/2008 */
/**
 * \internal
 * \brief Computes LOD for a scene
 * \param scene a scene
 * \param cam the camera used to compute LOD from
 * 
 * This function computes LOD (Level Of Detail) of the scene's models for the
 * given camera.
 */
static void SCE_Scene_ComputeLODs (SCE_SScene *scene, SCE_SCamera *cam)
{
    SCE_SListIterator *i;
    unsigned int ri = scene->ri;
    SCE_SSceneModelRenderInfos *infos = NULL;

    SCE_List_ForEach (i, scene->render[ri].models)
    {
        infos = SCE_List_GetData (i);
        if (infos->draw)
        {
            SCE_Model_UpdateLOD (infos->model, cam);
            /* UpdateLOD() can disable the render of the model */
            if (!SCE_Model_IsActived (infos->model))
                infos->draw = SCE_FALSE;
        }
    }
}

/* ajoute le 07/07/2008 */
/* revise le 16/09/2008 */
/**
 * \brief Prepares the active scene's renderer for rendering
 * \param scene a scene
 * \param camera the camera used for the render, can be NULL then a default
 *        camera (the same one that SCE_Camera_Create() returns) is used
 * \param rendertarget the render target
 * \param cubeface if the render target is a cubemap, it is used to determine
 *        which face is the render target. Otherwise, if the render target is
 *        not a cubmap, this parameter is useless.
 *        This can be one of SCE_TEX_POSX, SCE_TEX_NEGX, SCE_TEX_POSY,
 *        SCE_TEX_NEGY, SCE_TEX_POSZ or SCE_TEX_NEGZ.
 * 
 * This function prepares the given scene's active renderer for render. It
 * assignes a camera and a render target to the active scene's renderer.
 * 
 * \note cubeface is assigned to the active scene's renderer even if the render
 *       target is not a cubemap.
 * 
 * \see SCE_Texture_RenderTo()
 */
void SCE_Scene_Update (SCE_SScene *scene, SCE_SCamera *camera,
                       SCE_STexture *rendertarget, SCEuint cubeface)
{
    unsigned int ri = scene->ri;

    /* assignation des variables "d'environnement" */
    scene->render[ri].rendertarget = rendertarget;
    scene->render[ri].cubeface = cubeface;
    scene->render[ri].camera = (camera ? camera : default_camera);

    /* mise a jour des noeuds de la scene */
    SCE_Node_UpdateRecursive (scene->rootnode);

    SCE_Scene_SelectActivedModels (scene);

    /* --- gestion du frustum culling --- */
    if (scene->render[ri].states.frustum_culling)
        SCE_Scene_SelectInFrustumModels (scene, camera);

    /* --- determination du LOD --- */
    if (scene->render[ri].states.lod)
        SCE_Scene_ComputeLODs (scene, camera);

    /* --- calcul de toutes les lumieres a portee des objets --- */
    if (scene->render[ri].states.lighting)
        SCE_Scene_MakeLightsLists (scene);
}

/* ajoute le 10/07/2008 */
/**
 * \brief Switch up scene's active renderer
 * \param scene a scene
 * 
 * This functions sets the active renderer to the next one.
 */
void SCE_Scene_PushUpdate (SCE_SScene *scene)
{
    if (scene->ri + 1 >= SCE_MAX_RENDER_INFOS)
    {
        Logger_PrintMsg ("array overflow in updates' push list of the scene");
        return;
    }
    scene->ri++;
}
/**
 * \brief Switch down scene's active renderer
 * \param scene a scene
 * 
 * This functions sets the active renderer to the previous one.
 */
/* ajoute le 10/07/2008 */
void SCE_Scene_PopUpdate (SCE_SScene *scene)
{
    if (scene->ri > 0)
        scene->ri--;
}

/* ajoute le 08/07/2008 */
static void SCE_Scene_ActivateLights (SCE_SSceneModelRenderInfos *infos)
{
    unsigned int i;
    for (i=0; i < infos->num_lights; i++)
        SCE_Light_Use (infos->lights[i]);
}


/* ajoute le 03/04/2008 */
/* TODO: da shit func */
/**
 * \internal
 * 
 * \deprecated
 */
void SCE_Scene_PrepareRendering (SCE_SScene *scene)
{
    if (scene->render[scene->ri].states.use_skybox)
        SCE_CClear (GL_DEPTH_BUFFER_BIT);
    else
        SCE_CClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

/* ajoute le 19/03/2008 */
/* revise le 15/07/2008 */
/**
 * \brief Renders a scene into a render target
 * \param scene a scene
 * \param cam the camera used for the render or NULL to keep the current one.
 * \param rendertarget the render target or NULL to keep teh current one. If
 *        both this parameter and the current render targed are NULL, the
 *        default OpenGL's render buffer will be used as the render target.
 * \param cubeface see SCE_Scene_Update()'s cubeface parameter.
 * 
 * \see SCE_Scene_Update()
 */
void SCE_Scene_Render (SCE_SScene *scene, SCE_SCamera *cam,
                       SCE_STexture *rendertarget, int cubeface)
{
    SCE_SListIterator *i = NULL;
    SCE_SSceneModelRenderInfos *infos = NULL;
    int ri = scene->ri;

    if (!cam)
        cam = scene->render[ri].camera;
    if (!rendertarget)
        rendertarget = scene->render[ri].rendertarget;
    if (cubeface < 0)
        cubeface = scene->render[ri].cubeface;

    SCE_Scene_PrepareRendering (scene);

    /* mise en place du render target */
    SCE_Texture_RenderTo (rendertarget, cubeface);

    /* activation de la camera et mise en place des matrices */
    SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
    SCE_CPushMatrix ();
    SCE_Camera_Use (cam);

    /* assignation de la matrice camera au gestionnaire de lumieres */
    if (scene->render[ri].states.lighting)
    {
        SCE_Light_ActivateLighting (SCE_TRUE);
        SCE_Light_SetCameraMatrix (SCE_Camera_GetFinalView (cam));
    }
    else
        SCE_Light_ActivateLighting (SCE_FALSE);

    /* pour chaque modele de la scene */
    SCE_List_ForEach (i, scene->render[ri].models)
    {
        infos = SCE_List_GetData (i);
        if (infos->draw)
        {
            SCE_Scene_ActivateLights (infos);
            SCE_Model_Render (infos->model);
            SCE_Light_Use (NULL);
        }
    }

    SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
    SCE_CPopMatrix ();

    if (rendertarget)
        SCE_Texture_RenderTo (NULL, 0);

    SCE_CFlush ();
}

/** @} */
