/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 9 juillet 2007
   derniere modification : 05/04/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEResources.h>
#include <SCE/interface/SCESkybox.h>

static int num_skybox = 0;

/* ajoute le 09/07/2007 */
/* revise le 05/04/2008 */
SCE_SModel* SCE_Skybox_Create (float l)
{
    char resname[32] = {"\0"};
    SCE_SModel *mdl = SCE_Model_Create ();
    if (!mdl)
        Logger_LogSrc ();
    else
    {
        SCE_SMesh *mesh = NULL;
        SCE_TVector3 origin;
        
        origin[0] = -l/2.;
        origin[1] = -l/2.;
        origin[2] = -l/2.;
        
        mesh = SCE_Mesh_CreateCube (origin, l, l, l);
        if (mesh)
        {
            SCE_SMesh **meshs = SCE_malloc (2 * sizeof *meshs);
            if (!meshs)
            {
                SCE_Model_Delete (mdl);
                Logger_LogSrc ();
                return NULL;
            }
            meshs[0] = mesh;
            meshs[1] = NULL;
            SCE_Model_SetMeshs (mdl, meshs, 1);
            SCE_Model_DisableBackfaceCulling (mdl);
            sprintf (resname, "sce_skybox_number_%d", num_skybox);
            SCE_Resource_Add (resname, meshs);
            sprintf (resname, "sce_skybox_number_%d_lod", num_skybox);
            SCE_Resource_Add (resname, mesh);
            num_skybox++;
            SCE_Material_SetColor (SCE_Model_GetMaterial (mdl),
                                   SCE_DIFFUSE_COLOR, 1., 1., 1., 1.);
            SCE_Material_SetColor (SCE_Model_GetMaterial (mdl),
                                   SCE_AMBIENT_COLOR, 1., 1., 1., 1.);
        }
        else
        {
            SCE_Model_Delete (mdl), mdl = NULL;
            Logger_LogSrc ();
        }
    }
    return mdl;
}

/* ajoute le 05/04/2008 */
int SCE_Skybox_SetTextureFromFile (SCE_SModel *mdl, int type, int mipmap, ...)
{
    va_list args;
    SCE_STexture *tex = NULL;
    SCE_btstart ();
    va_start (args, mipmap);
    tex = SCE_Texture_LoadArg (type, 0, 0, 0, args);
    va_end (args);
    if (SCE_Texture_Build (tex, mipmap) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    SCE_btend ();
    return SCE_Skybox_SetTexture (mdl, tex, SCE_TRUE);
}
/* ajoute le 05/04/2008 */
int SCE_Skybox_SetTexture (SCE_SModel *mdl, SCE_STexture *tex, int canfree)
{
    SCEvertices *v = NULL;
    SCEvertices tex2d[] =
    {
        0., 0.,  1., 0.,  1., 1.,  0., 1.,
        0., 0.,  1., 0.,  1., 1.,  0., 1.,
        0., 0.,  1., 0.,  1., 1.,  0., 1.,
        0., 0.,  1., 0.,  1., 1.,  0., 1.,
        0., 0.,  1., 0.,  1., 1.,  0., 1.,
        0., 0.,  1., 0.,  1., 1.,  0., 1.,
    };
    SCE_TVector3 texcube[24];
    SCE_SMesh *mesh = *SCE_Model_GetMeshs (mdl);
    int i;
    int type = SCE_Texture_GetCType (tex);

    switch (type)
    {
    case SCE_TEX_2D:
        /* la meme texture sera appliquee sur tous les cotes de la skybox */
        if (SCE_Mesh_AddVerticesDup (mesh, 0, SCE_TEXCOORD0,
                                     SCE_VERTICES_TYPE, 2, 24, tex2d) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        if (SCE_Mesh_Build (mesh) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        break;
    case SCE_TEX_CUBE:
        /* une texture cubemap sera appliquee a la skybox */
        v = SCE_Mesh_LocateData (mesh, SCE_POSITION, NULL)->data;
        for (i=0; i<24; i++)
        {
            SCE_Vector3_Copy (texcube[i*3], &v[i*3]);
            SCE_Vector3_Normalize (texcube[i*3]);
        }
        if (SCE_Mesh_AddVerticesDup (mesh, 0, SCE_TEXCOORD0, SCE_FLOAT,
                                     3, 24, texcube) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        if (SCE_Mesh_Build (mesh) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        break;
#ifdef SCE_DEBUG
    default:
        Logger_Log (SCE_INVALID_OPERATION);
        /* TODO: da messsage */
        Logger_LogMsg ("texture type unsupported for the skybox mapping");
#endif
    }
    if (SCE_Material_AddTexture (SCE_Model_GetMaterial (mdl),
                                 0, tex, canfree) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 05/04/2008 */
void SCE_Skybox_SetTexModel (SCE_SModel *mdl, int *texmdl, int w, int h)
{
    /* useful function but not written § :D */
}
